/*  CPfront 1.0 -48 */
#include "SYSTEM.h"
#include "Kernel.h"
#include "C99types.h"
#include "Console.h"
#include "Files.h"
#include "Strings.h"
#include "Services.h"
#include "Log.h"

typedef
	struct DswLog_Hook__rec *DswLog_Hook;

typedef
	struct DswLog_Hook__rec {
		char _prvt0[1];
	} DswLog_Hook__rec;

static void DswLog_Hook_Beep (DswLog_Hook log);
static void DswLog_Hook_Bool (DswLog_Hook log, BOOLEAN x);
static void DswLog_Hook_Char (DswLog_Hook log, CHAR ch);
static void DswLog_Hook_ClearBuf (DswLog_Hook log);
static void DswLog_Hook_FlushBuf (DswLog_Hook log);
static BOOLEAN DswLog_Hook_Guard (DswLog_Hook log, ANYPTR o);
static void DswLog_Hook_Int (DswLog_Hook log, INTEGER n);
static void DswLog_Hook_IntForm (DswLog_Hook log, INTEGER x, INTEGER base, INTEGER minWidth, CHAR fillCh, \
BOOLEAN showBase);
static void DswLog_Hook_Ln (DswLog_Hook log);
static void DswLog_Hook_Para (DswLog_Hook log);
static void DswLog_Hook_ParamMsg (DswLog_Hook log, CHAR *s, INTEGER s__len, CHAR *p0, INTEGER p0__len, \
CHAR *p1, INTEGER p1__len, CHAR *p2, INTEGER p2__len);
static void DswLog_Hook_Real (DswLog_Hook log, REAL x);
static void DswLog_Hook_RealForm (DswLog_Hook log, REAL x, INTEGER precision, INTEGER minW, INTEGER expW, \
CHAR fillCh);
static void DswLog_Hook_Set (DswLog_Hook log, SET x);
static void DswLog_Hook_String (DswLog_Hook log, CHAR *str, INTEGER str__len);
static void DswLog_Hook_Tab (DswLog_Hook log);
static void DswLog_Hook_View (DswLog_Hook log, ANYPTR v);
static void DswLog_Hook_ViewForm (DswLog_Hook log, ANYPTR v, INTEGER w, INTEGER h);


export BOOLEAN DswLog_beep, DswLog_console, DswLog_view;
export Files_Writer DswLog_writer;

export INTEGER DswLog_Hook__rec__desc[];
export SYSTEM_TYPEDESC *DswLog_Hook__rec__typ = (SYSTEM_TYPEDESC*)(DswLog_Hook__rec__desc + 20);

static void DswLog_FileByte (SHORTCHAR ch);
static void DswLog_Flush (void);
static void DswLog_Init (void);
static void DswLog_WriteChar (CHAR ch);
static void DswLog_WriteString (CHAR *str, INTEGER str__len);

export void DswLog__reg();
export void DswLog__body();
export struct SYSTEM_MODDESC DswLog__desc;


static void DswLog_Flush (void)
{
	__ENTER("DswLog.Flush");
	if (DswLog_writer != NIL) {
		__Files_File_Flush(__Files_Writer_Base(DswLog_writer));
	}
	__EXIT;
}

static void DswLog_FileByte (SHORTCHAR ch)
{
	__ENTER("DswLog.FileByte");
	__Files_Writer_WriteByte(DswLog_writer, (BYTE)ch);
	if (Log_synch) {
		DswLog_Flush();
	}
	__EXIT;
}

typedef
	BYTE __23[4];

static void DswLog_WriteChar (CHAR ch)
{
	INTEGER i, res;
	CHAR str[2];
	SHORTCHAR sstr[4];
	__ENTER("DswLog.WriteChar");
	if (DswLog_console) {
		Console_WriteChar(ch);
	}
	if (DswLog_writer != NIL) {
		if (ch >= 0 && ch <= 127) {
			__Files_Writer_WriteByte(DswLog_writer, (BYTE)ch);
		} else {
			str[0] = ch;
			str[1] = 0;
			Strings_StringToUtf8((void*)str, 2, (void*)sstr, 4, &res);
			__Files_Writer_WriteBytes(DswLog_writer, (__23*)sstr, 4, 0, __STRLENS(sstr) - 1);
		}
		if (Log_synch) {
			DswLog_Flush();
		}
	}
	__EXIT;
}

typedef
	struct BYTE_ARRAY *__26;

static void DswLog_WriteString (CHAR *str, INTEGER str__len)
{
	INTEGER i, len, res;
	struct SHORTCHAR_ARRAY *sstr = NIL;
	INTEGER _for__25;
	__ENTER("DswLog.WriteString");
	__DUP(str, str__len);
	if (DswLog_console) {
		Console_WriteStr((void*)str, str__len);
	}
	if (DswLog_writer != NIL) {
		sstr = __NEWARR(1, __STRLEN(str) * 3 + 1);
		Strings_StringToUtf8((void*)str, str__len, (void*)sstr->data, sstr->len[0], &res);
		_for__25 = __STRLENS(sstr->data) - 1;
		i = 0;
		while (i <= _for__25) {
			__Files_Writer_WriteByte(DswLog_writer, (BYTE)(sstr->data)[__X(i, sstr->len[0])]);
			i += 1;
		}
		if (Log_synch) {
			DswLog_Flush();
		}
	}
	__DEL(str);
	__EXIT;
}

static BOOLEAN DswLog_Hook_Guard (DswLog_Hook log, ANYPTR o)
{
	__ENTER("DswLog.Hook.Guard");
	__EXIT;
	return 0;
}

static void DswLog_Hook_ClearBuf (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.ClearBuf");
	__EXIT;
}

static void DswLog_Hook_FlushBuf (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.FlushBuf");
	DswLog_Flush();
	__EXIT;
}

static void DswLog_Hook_Beep (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.Beep");
	if (DswLog_beep) {
		Kernel_Beep();
	}
	__EXIT;
}

static void DswLog_Hook_Char (DswLog_Hook log, CHAR ch)
{
	__ENTER("DswLog.Hook.Char");
	DswLog_WriteChar(ch);
	__EXIT;
}

static void DswLog_Hook_Int (DswLog_Hook log, INTEGER n)
{
	CHAR str[16];
	__ENTER("DswLog.Hook.Int");
	if (DswLog_console || DswLog_writer != NIL) {
		Strings_IntToString(n, (void*)str, 16);
		DswLog_WriteChar((CHAR)L' ');
		DswLog_WriteString(str, 16);
	}
	__EXIT;
}

static void DswLog_Hook_Real (DswLog_Hook log, REAL x)
{
	CHAR str[256];
	__ENTER("DswLog.Hook.Real");
	if (DswLog_console || DswLog_writer != NIL) {
		Strings_RealToString(x, (void*)str, 256);
		DswLog_WriteChar((CHAR)L' ');
		DswLog_WriteString(str, 256);
	}
	__EXIT;
}

static void DswLog_Hook_String (DswLog_Hook log, CHAR *str, INTEGER str__len)
{
	__ENTER("DswLog.Hook.String");
	DswLog_WriteString(str, str__len);
	__EXIT;
}

static void DswLog_Hook_Bool (DswLog_Hook log, BOOLEAN x)
{
	__ENTER("DswLog.Hook.Bool");
	if (x) {
		DswLog_WriteString((CHAR*)((CHAR[]){32, 36, 84, 82, 85, 69, 0}), 7);
	} else {
		DswLog_WriteString((CHAR*)((CHAR[]){32, 36, 70, 65, 76, 83, 69, 0}), 8);
	}
	__EXIT;
}

static void DswLog_Hook_Set (DswLog_Hook log, SET x)
{
	CHAR str[256];
	__ENTER("DswLog.Hook.Set");
	if (DswLog_console || DswLog_writer != NIL) {
		Strings_SetToString(x, (void*)str, 256);
		DswLog_WriteChar((CHAR)L' ');
		DswLog_WriteString(str, 256);
	}
	__EXIT;
}

static void DswLog_Hook_IntForm (DswLog_Hook log, INTEGER x, INTEGER base, INTEGER minWidth, CHAR fillCh, \
BOOLEAN showBase)
{
	CHAR str[40];
	__ENTER("DswLog.Hook.IntForm");
	if (DswLog_console || DswLog_writer != NIL) {
		Strings_IntToStringForm(x, base, minWidth, fillCh, showBase, (void*)str, 40);
		DswLog_WriteString(str, 40);
	}
	__EXIT;
}

static void DswLog_Hook_RealForm (DswLog_Hook log, REAL x, INTEGER precision, INTEGER minW, INTEGER expW, \
CHAR fillCh)
{
	CHAR str[256];
	__ENTER("DswLog.Hook.RealForm");
	if (DswLog_console || DswLog_writer != NIL) {
		Strings_RealToStringForm(x, precision, minW, expW, fillCh, (void*)str, 256);
		DswLog_WriteString(str, 256);
	}
	__EXIT;
}

static void DswLog_Hook_Tab (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.Tab");
	DswLog_WriteChar(9);
	__EXIT;
}

static void DswLog_Hook_Ln (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.Ln");
	if (DswLog_console) {
		Console_WriteLn();
	}
	if (DswLog_writer != NIL) {
		DswLog_FileByte(10);
	}
	__EXIT;
}

static void DswLog_Hook_Para (DswLog_Hook log)
{
	__ENTER("DswLog.Hook.Para");
	if (DswLog_console) {
		Console_WriteLn();
	}
	if (DswLog_writer != NIL) {
		DswLog_FileByte(10);
	}
	__EXIT;
}

static void DswLog_Hook_View (DswLog_Hook log, ANYPTR v)
{
	CHAR str[256];
	__ENTER("DswLog.Hook.View");
	if (DswLog_console || DswLog_writer != NIL) {
		if (v == NIL) {
			__MOVE(((CHAR[]){78, 73, 76, 0}), str, 8);
		} else {
			Services_GetTypeName(v, __TYPEOF(v), (void*)str, 256);
		}
		DswLog_WriteChar((CHAR)L'[');
		DswLog_WriteString(str, 256);
		DswLog_WriteChar((CHAR)L']');
	}
	__EXIT;
}

static void DswLog_Hook_ViewForm (DswLog_Hook log, ANYPTR v, INTEGER w, INTEGER h)
{
	CHAR str[256];
	__ENTER("DswLog.Hook.ViewForm");
	if (DswLog_console || DswLog_writer != NIL) {
		if (v == NIL) {
			__MOVE(((CHAR[]){78, 73, 76, 0}), str, 8);
		} else {
			Services_GetTypeName(v, __TYPEOF(v), (void*)str, 256);
		}
		DswLog_WriteString((CHAR*)((CHAR[]){91, 91, 0}), 3);
		DswLog_WriteString(str, 256);
		DswLog_WriteChar((CHAR)L':');
		Strings_IntToString(w, (void*)str, 256);
		DswLog_WriteString(str, 256);
		DswLog_WriteChar((CHAR)L':');
		Strings_IntToString(h, (void*)str, 256);
		DswLog_WriteString(str, 256);
		DswLog_WriteString((CHAR*)((CHAR[]){93, 93, 0}), 3);
	}
	__EXIT;
}

static void DswLog_Hook_ParamMsg (DswLog_Hook log, CHAR *s, INTEGER s__len, CHAR *p0, INTEGER p0__len, \
CHAR *p1, INTEGER p1__len, CHAR *p2, INTEGER p2__len)
{
	INTEGER i;
	__ENTER("DswLog.Hook.ParamMsg");
	if (DswLog_console || DswLog_writer != NIL) {
		i = 0;
		if ((s[0] == (CHAR)L'#' && s[__X(1, s__len)] != (CHAR)L':') && s[__X(1, s__len)] != 0) {
			i = 1;
			while (s[__X(i, s__len)] != (CHAR)L':' && s[__X(i, s__len)] != 0) {
				i += 1;
			}
			if (s[__X(i, s__len)] != (CHAR)L':') {
				i = 0;
			}
		}
		while (s[__X(i, s__len)] != 0) {
			if (s[__X(i, s__len)] == (CHAR)L'^') {
				switch (s[__X(i + 1, s__len)]) {
					case (CHAR)'0': 
						DswLog_WriteString(p0, p0__len);
						break;
					case (CHAR)'1': 
						DswLog_WriteString(p1, p1__len);
						break;
					case (CHAR)'2': 
						DswLog_WriteString(p2, p2__len);
						break;
					default: 
						DswLog_WriteChar((CHAR)L'^');
						break;
				}
				i += 2;
			} else {
				DswLog_WriteChar(s[__X(i, s__len)]);
				i += 1;
			}
		}
	}
	__EXIT;
}

static void DswLog_Init (void)
{
	DswLog_Hook logHook = NIL;
	__ENTER("DswLog.Init");
	DswLog_beep = 1;
	DswLog_console = 1;
	DswLog_view = 1;
	DswLog_writer = NIL;
	logHook = __NEW(DswLog_Hook__rec__typ);
	Log_SetHook((void*)logHook);
	__EXIT;
}

static INTEGER DswLog_Hook__rec__flds[] = {
	0, 
};
export INTEGER DswLog_Hook__rec__desc[] = {
	-1, 
	(INTEGER)DswLog_Hook_ViewForm, 
	(INTEGER)DswLog_Hook_View, 
	(INTEGER)DswLog_Hook_Tab, 
	(INTEGER)DswLog_Hook_String, 
	(INTEGER)DswLog_Hook_Set, 
	(INTEGER)DswLog_Hook_RealForm, 
	(INTEGER)DswLog_Hook_Real, 
	(INTEGER)DswLog_Hook_ParamMsg, 
	(INTEGER)DswLog_Hook_Para, 
	(INTEGER)DswLog_Hook_Ln, 
	(INTEGER)DswLog_Hook_IntForm, 
	(INTEGER)DswLog_Hook_Int, 
	(INTEGER)DswLog_Hook_Guard, 
	(INTEGER)DswLog_Hook_FlushBuf, 
	(INTEGER)DswLog_Hook_ClearBuf, 
	(INTEGER)DswLog_Hook_Char, 
	(INTEGER)DswLog_Hook_Bool, 
	(INTEGER)DswLog_Hook_Beep, 
	0, 
	1,
	(INTEGER)&DswLog__desc,
	1<<8 | 0x21,
	(INTEGER)(Kernel_Hook__rec__desc + 2),
	(INTEGER)(Log_Hook__rec__desc + 20),
	(INTEGER)(DswLog_Hook__rec__desc + 20),
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	(INTEGER)DswLog_Hook__rec__flds, 
	-4
};
static SYSTEM_MODDESC *DswLog__imp[] = {
	&Console__desc,
	&Files__desc,
	&Kernel__desc,
	&Log__desc,
	&Services__desc,
	&Strings__desc,
};
static INTEGER DswLog__exp[] = {
	10,
	0x227d6f36, (INTEGER)DswLog_FileByte, 7<<8 | 0x14, 0,
	0x227d6f36, (INTEGER)DswLog_Flush, 16<<8 | 0x14, 0,
	0x83cd1ac9, 0x968c1b82, 1<<8 | 0x12, (INTEGER)(DswLog_Hook__rec__desc + 20),
	0x227d6f36, (INTEGER)DswLog_Init, 22<<8 | 0x14, 0,
	0x227d6f36, (INTEGER)DswLog_WriteChar, 27<<8 | 0x14, 0,
	0x227d6f36, (INTEGER)DswLog_WriteString, 37<<8 | 0x14, 0,
	0x44ec5d5b, (INTEGER)&DswLog_beep, 49<<8 | 0x43, 1,
	0x44ec5d5b, (INTEGER)&DswLog_console, 54<<8 | 0x43, 1,
	0x44ec5d5b, (INTEGER)&DswLog_view, 62<<8 | 0x43, 1,
	0x69494078, (INTEGER)&DswLog_writer, 67<<8 | 0x43, (INTEGER)Files_Writer__desc,
};
static char DswLog__names[] = {
	0,
	'H','o','o','k','^',0,
	'F','i','l','e','B','y','t','e',0,
	'F','l','u','s','h',0,
	'I','n','i','t',0,
	'W','r','i','t','e','C','h','a','r',0,
	'W','r','i','t','e','S','t','r','i','n','g',0,
	'b','e','e','p',0,
	'c','o','n','s','o','l','e',0,
	'v','i','e','w',0,
	'w','r','i','t','e','r',0,
};
static INTEGER DswLog__ptrs[] = {
	(INTEGER)&DswLog_writer + 0, 
};
struct SYSTEM_MODDESC DswLog__desc = {
	0, 13, 0, /* next, opts, refcnt */ 
	{2019, 6, 9, 21, 27, 19}, /* compTime */ 
	{0, 0, 0, 0, 0, 0}, /* loadTime */ 
	DswLog__body,
	0,
	6, /* nofimps */ 
	1, /* nofptrs */ 
	0, 0, 0, 0, 0, 0, 0, 0, /* csize..varBase */ 
	DswLog__names,
	DswLog__ptrs,
	DswLog__imp,
	(SYSTEM_DIRECTORY*)DswLog__exp,
	"DswLog"
};

export void DswLog__reg()
{
	__BEGREG(DswLog__desc);
	Kernel__reg();
	Console__reg();
	Files__reg();
	Strings__reg();
	Services__reg();
	Log__reg();
	__REGMOD(DswLog__desc);
	DswLog_Hook__rec__desc[19] = Log_Hook__rec__desc[19];
	__ENDREG;
}

export void DswLog__body()
{
	__BEGBODY(DswLog__desc);
	Kernel__body();
	Console__body();
	Files__body();
	Strings__body();
	Services__body();
	Log__body();
	DswLog_Init();
	__ENDBODY;
}
