#! /bin/sh

set -e

THIS="$(dirname "$(readlink -f "$0")")"
OUT="$THIS/stage0/i486"

###^^^^^^^^^^^^^^^^^^###
### Check BBCP files ###
###__________________###

mkdir -p "$THIS/stage0"
if ! [ -e "$THIS/stage0/bbcp" ]; then
	echo "Please, clone bbcp repo:"
	echo "->  git clone --depth=1 -b crux https://git.deadsoftware.ru/bbcp.git '$(realpath --relative-to="$PWD" "$THIS/stage0/bbcp")'"
	echo "Than retry building using make-stage0.sh"
	exit 1
fi
cd "$THIS/stage0/bbcp/BlackBox"

###^^^^^^^^^^^^^^^^^^^###
### Build BBCP itself ###
###___________________###

./switch-target Linux Interp
./build

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Copy minimal bbdsw sources ###
###____________________________###

rm -rf "$OUT"
mkdir -p "$OUT"
cp -r -- \
	"$THIS/src/generic/"* \
	"$THIS/src/posix/"* \
	"$THIS/src/i486/generic/"* \
	"$THIS/src/i486/posix/"* \
	"$THIS/src/i486/linux/"* \
	"$OUT"

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile Stage0 Compiler & Linker ###
###__________________________________###

./blackbox <<!

Kernel.LoadMod('DevCompiler')
Kernel.LoadMod('ConsCompiler')
Kernel.LoadMod('Dev2Linker1')
Kernel.LoadMod('Dev2LnkChmod')

HostFiles.SetRootDir('$OUT')

ConsCompiler.Compile('C99/Mod', 'types.cp')
ConsCompiler.Compile('C99/Mod', 'sys_types.cp')
ConsCompiler.Compile('C99/Mod', 'stdlib.cp')
ConsCompiler.Compile('C99/Mod', 'stdio.cp')
ConsCompiler.Compile('C99/Mod', 'unistd.cp')
ConsCompiler.Compile('C99/Mod', 'dirent.cp')
ConsCompiler.Compile('C99/Mod', 'locale.cp')
ConsCompiler.Compile('C99/Mod', 'time.cp')
ConsCompiler.Compile('C99/Mod', 'sys_stat.cp')
ConsCompiler.Compile('C99/Mod', 'fcntl.cp')
ConsCompiler.Compile('C99/Mod', 'errno.cp')
ConsCompiler.Compile('C99/Mod', 'iconv.cp')
ConsCompiler.Compile('C99/Mod', 'wctype.cp')
ConsCompiler.Compile('C99/Mod', 'sys_mman.cp')
ConsCompiler.Compile('C99/Mod', 'dlfcn.cp')
ConsCompiler.Compile('C99/Mod', 'signal.cp')
ConsCompiler.Compile('C99/Mod', 'setjmp.cp')
ConsCompiler.Compile('C99/Mod', 'libgen.cp')
ConsCompiler.Compile('C99/Mod', 'macro.cp')

ConsCompiler.Compile('System/Mod', 'Kernel.cp')
DevCompiler.CompileThis Console Dates Files Int Long Math SMath Strings Log Services

ConsCompiler.Compile('Host/Mod', 'Lang.cp')
ConsCompiler.Compile('Host/Mod', 'Dates.cp')
ConsCompiler.Compile('Host/Mod', 'Console.cp')
ConsCompiler.Compile('Host/Mod', 'Files.cp')

DevCompiler.CompileThis \
	DswLog DswDebug \
	Dev2LnkBase Dev2LnkChmod Dev2LnkLoad Dev2LnkWriteElf Dev2LnkWriteElfStatic Dev2LnkWritePe
ConsCompiler.Compile('Dsw/Mod','Linker486Main.cp')
Dev2Linker1.LinkElf Linux cpl486 := Kernel$+ \
	C99types C99macro \
	Kernel \
	Console Files Math Strings Services Log Int Long \
	HostLang HostConsole HostFiles DswLog DswDebug \
	Dev2LnkBase Dev2LnkChmod Dev2LnkLoad Dev2LnkWriteElf \
	Dev2LnkWriteElfStatic Dev2LnkWritePe \
	DswLinker486Main

ConsCompiler.Compile('Dev/Mod','CPM.cp')
DevCompiler.CompileThis DevCPT DevCPS DevCPB DevCPP DevCPE DevCPH DevCPL486 DevCPC486 DevCPV486
ConsCompiler.Compile('Dsw/Mod','Documents.cp')
ConsCompiler.Compile('Dsw/Mod','Compiler486Main.cp')
Dev2Linker1.LinkElf Linux cpc486 := Kernel$+ \
	C99types C99macro \
	Kernel \
	Console Files Math Strings Services Log Int Long \
	HostLang HostConsole HostFiles HostDates DswLog DswDebug \
	DevCPM DevCPT DevCPS DevCPB DevCPP DevCPE DevCPH DevCPL486 DevCPC486 DevCPV486 \
	DswDocuments DswCompiler486Main

Kernel.Quit(0)

!

cd "$OUT"
chmod a+x cpc486 cpl486
