#! /bin/sh

set -e

THIS="$(dirname "$(readlink -f "$0")")"
OUT="$THIS/stage2/i486"

cpc() {
	"$THIS/stage1/i486/cpc486" -legacy "$@"
}

cpl() {
	"$THIS/stage1/i486/cpl486" "$@"
}

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Prepare bbdsw sources for LINUX/i486 ###
###______________________________________###

rm -rf "$OUT"
mkdir -p "$OUT"
cp -r -- \
	"$THIS/src/generic/"* \
	"$THIS/src/posix/"* \
	"$THIS/src/i486/generic/"* \
	"$THIS/src/i486/posix/"* \
	"$THIS/src/i486/linux/"* \
	"$OUT"
cd "$OUT"

###^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile POSIX bindings ###
###________________________###

cpc C99/Mod/types.cp \
	C99/Mod/sys_types.cp \
	C99/Mod/stdlib.cp C99/Mod/stdio.cp C99/Mod/unistd.cp \
	C99/Mod/dirent.cp C99/Mod/locale.cp C99/Mod/time.cp \
	C99/Mod/sys_stat.cp C99/Mod/fcntl.cp C99/Mod/errno.cp \
	C99/Mod/iconv.cp C99/Mod/wctype.cp C99/Mod/sys_mman.cp \
	C99/Mod/dlfcn.cp C99/Mod/signal.cp C99/Mod/setjmp.cp \
	C99/Mod/libgen.cp \
	C99/Mod/macro.cp

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile BlackBox Framework ###
###____________________________###


cpc System/Mod/Int.odc System/Mod/Long.odc \
	System/Mod/Math.odc System/Mod/SMath.odc System/Mod/Kernel.cp \
	System/Mod/Console.odc System/Mod/Files.odc System/Mod/Dates.odc \
	System/Mod/Log.odc System/Mod/Strings.odc System/Mod/Services.odc \
	System/Mod/Integers.odc
mv Code Sym System

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile Linux Host subsystem ###
###______________________________###

cpc Host/Mod/Lang.cp Host/Mod/Dates.cp Host/Mod/Console.cp Host/Mod/Files.cp

###^^^^^^^^^^^^^^^^^^^^^^^###
### Compile Dev subsystem ###
###_______________________###

cpc Dev/Mod/CPM.cp Dev/Mod/CPT.odc Dev/Mod/CPS.odc Dev/Mod/CPB.odc \
	Dev/Mod/CPP.odc Dev/Mod/CPE.odc Dev/Mod/CPH.odc Dev/Mod/CPL486.odc \
	Dev/Mod/CPC486.odc Dev/Mod/CPV486.odc

###^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile Dev2 subsystem ###
###________________________###

cpc Dev2/Mod/LnkBase.odc Dev2/Mod/LnkChmod.odc Dev2/Mod/LnkLoad.odc \
	Dev2/Mod/LnkWriteElf.odc Dev2/Mod/LnkWriteElfStatic.odc \
	Dev2/Mod/LnkWritePe.odc

###^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile CPfront subsystem ###
###___________________________###

cpc CPfront/Mod/CPG.odc CPfront/Mod/CPC.odc CPfront/Mod/CPV.odc

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Compile bbdsw-specific modules ###
###________________________________###

cpc Dsw/Mod/Documents.cp Dsw/Mod/Log.odc Dsw/Mod/Debug.odc Dsw/Mod/Compiler486Main.cp \
	Dsw/Mod/CompilerCPfrontMain.cp Dsw/Mod/Linker486Main.cp

###^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^###
### Link Standalone Component Pascl Compiler & Dev2 Linker ###
###________________________________________________________###

cpl -os linux -kernel Kernel -main Kernel -legacycodedir . -o cpc486 \
	Long \
	C99types C99macro \
	Kernel Console Files Dates Math Strings Services Log \
	HostLang HostConsole HostFiles HostDates DswLog DswDebug \
	DevCPM DevCPT DevCPS DevCPB DevCPP DevCPE DevCPH \
	DevCPL486 DevCPC486 DevCPV486 \
	DswDocuments DswCompiler486Main

cpl -os linux -kernel Kernel -main Kernel -legacycodedir . -o cpfront \
	Long \
	C99types C99macro \
	Kernel Console Files Dates Math Strings Services Log \
	HostLang HostConsole HostFiles HostDates DswLog DswDebug \
	DevCPM DevCPT DevCPS DevCPB DevCPP DevCPE DevCPH \
	CPfrontCPG CPfrontCPC CPfrontCPV\
	DswDocuments DswCompilerCPfrontMain

cpl -os linux -kernel Kernel -main Kernel -legacycodedir . -o cpl486 \
	Long \
	C99types C99macro \
	Kernel Console Files Math Strings Services Log \
	HostLang HostConsole HostFiles DswLog DswDebug \
	Dev2LnkBase Dev2LnkChmod Dev2LnkLoad Dev2LnkWriteElf \
	Dev2LnkWriteElfStatic Dev2LnkWritePe \
	DswLinker486Main

#chmod a+x "$OUT/cpfront" "$OUT/cpc486" "$OUT/cpl486"
